/* Gauntlet, the beginning of *the* multiuser game */
/* Future plans: graphic, multi-user abilities, nethack treasure, etc */

#define G_BLANK 'C'
#define G_START 'x'
#define G_CHAR 'A'
#define G_WALL 'B'
#define G_MONSTER 'M'
#define G_BULLET '.'
#define MAXNROFPLAYERS 8
#define STARTMAX 100
#define MAX_BUF 256
#define MAX_X 64
#define MAX_Y 64
#define PLAYER 1
#define BULLET 2
#define MONSTER 3
#define TREASURE 4
#define obstacle(x,y) (x<0||x>mapx||y<0||y>mapy||map[x][y]!=G_BLANK)
#define UNDO op->speed++,flag=1
#define WINUPPER (-5)
#define WINLOWER 5
#define WINLEFT (-5)
#define WINRIGHT 5
#define MAX_NAME 16

#include <strings.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <signal.h>

#ifdef 0
#include "bitmaps.h"
#endif

typedef struct liv {
  int Str,Dex,Con,Wis,Cha,Int;
  int hp;   /* Hit Points.  When 0 is reached, the object will be deleted. */
  int exp;  /* Experience.  Killers gain 1/10. */
  int food; /* How much food in stomach.  When it reaches 0, you're dead. */
} living;
  
typedef struct obj {
  int x,y; /* Position */
  float speed,speed_left;
  int dirx,diry; /* Means the object is moving that way. */
  int type; /* PLAYER, BULLET, etc */
  char face;
  struct obj *next,*prev;
  struct pl *contr;
  living stats; /* I'll make this a pointer when I've made the
                       allocation system */
} object;

typedef struct pl {
  int last_refresh;
  Display *gdisp;
  Window win_game;
  Window win_stats;
  GC gc_game;
  GC gc_stats;
  XEvent gevent;
  KeySym gkey;
  XSizeHints gamehint;
  XSizeHints stathint;
  int gscreen;
  unsigned long gforeground,gbackground;
  Font font;
  object *ob; /* The object representing the player */
  char drawn[WINRIGHT-WINLEFT+1][WINLOWER-WINUPPER+1]; /* Last displayed */
  char name[MAX_NAME];
} player;

static char font_graphic[]="gauntlet";
static char font_text[]="8x13";
static int winwidth=WINRIGHT-WINLEFT+1;

player players[MAXNROFPLAYERS];
int nrofplayers; /* The current nr of players */

object objarray[STARTMAX]; /* All objects, allocated this way at first */
object *objects;      /* Pointer to the list of used objects */
object *free_objects; /* Pointer to the list of unused objects */

char name[]="Gauntlet";
char map[MAX_X][MAX_Y];
char omap[MAX_X][MAX_Y];
int mapx,mapy,startx,starty;

object *get_object();
object *get_player();
int gargc;
char **gargv;

init() {
  FILE *fp;
  int i,j;
  char buf[MAX_BUF];
  player *p;

/* Reading the map */
  if((fp=fopen("gmap","r"))==NULL) {
    fprintf(stderr,"Can't open gmap.\n");
    exit(1);
  }
  for(i=0;i<MAX_X;i++)
    for(j=0;j<MAX_Y;j++)
      omap[i][j]=G_BLANK;
  j=0,mapx=0,mapy=0,startx=2,starty=2;
  while(fgets(buf,MAX_BUF,fp)!=NULL) {
    for(i=0;i<MAX_X&&buf[i]!='\0'&&buf[i]!='\n';i++) {
      if(buf[i]==G_START)
        startx=i,starty=j,buf[i]=G_BLANK;
      omap[i][j]=buf[i];
    }
    if(i>mapx) mapx=--i;
    j++;
  }
  mapy=--j;

/* Initialize all objects: */
  free_objects=objarray;
  objects=NULL;
  objarray[0].prev=NULL,
  objarray[0].next=&objarray[1];
  for(i=1;i<STARTMAX-1;i++)
    objarray[i].next=&objarray[i+1],
    objarray[i].prev=&objarray[i-1];
  objarray[STARTMAX-1].next=NULL;
  objarray[STARTMAX-1].prev=&objarray[STARTMAX-2];

  for(i=1;i<gargc;i++) {
    p=&players[nrofplayers++];
    p->ob=get_player(p);
    strncpy(buf,gargv[i],MAX_BUF);
    if(strchr(buf,':')==NULL)
      strcat(buf,":0");
    get_main_display(p,buf);
    get_stats_display(p,buf);
  }
}

get_stats_display(player *p,char *name) {
  strncpy(p->name,name,MAX_NAME);
  p->stathint.x=200,p->stathint.y=200;
  p->stathint.width=300,p->stathint.height=100;
  p->stathint.flags=PPosition | PSize;
  p->win_stats=XCreateSimpleWindow(p->gdisp,DefaultRootWindow(p->gdisp),
      p->stathint.x,p->stathint.y,p->stathint.width,p->stathint.height,2,
      p->gbackground,p->gforeground);
  XSetStandardProperties(p->gdisp,p->win_stats,name,name,None,gargv,gargc,
                         &(p->stathint));
  p->gc_stats=XCreateGC(p->gdisp,p->win_stats,0,0);
  XSetForeground(p->gdisp,p->gc_stats,p->gbackground);
  XSetBackground(p->gdisp,p->gc_stats,p->gforeground);
  p->font=XLoadFont(p->gdisp,font_text);
  XSetFont(p->gdisp,p->gc_stats,p->font);
  XSelectInput(p->gdisp,p->win_stats,KeyPressMask|ExposureMask);
  XMapRaised(p->gdisp,p->win_stats);
}

get_main_display(player *p,char *display) {
  p->gdisp=XOpenDisplay(display);
  if(!p->gdisp) {
    fprintf(stderr,"Can't open display %s.\n",display);
    exit(1);
  }
  p->gscreen=DefaultScreen(p->gdisp);
  p->gbackground=BlackPixel(p->gdisp,p->gscreen);
  p->gforeground=WhitePixel(p->gdisp,p->gscreen);
  p->gamehint.x=300,p->gamehint.y=300;
  p->gamehint.width=350,p->gamehint.height=350;
  p->gamehint.flags=PPosition | PSize;
  p->win_game=XCreateSimpleWindow(p->gdisp,DefaultRootWindow(p->gdisp),
               p->gamehint.x,p->gamehint.y,
               p->gamehint.width,p->gamehint.height,2,
               p->gforeground,p->gbackground);
  XSetStandardProperties(p->gdisp,p->win_game,name,name,None,gargv,gargc,
                         &(p->gamehint));
  p->gc_game=XCreateGC(p->gdisp,p->win_game,0,0);
  XSetForeground(p->gdisp,p->gc_game,p->gbackground);
  XSetBackground(p->gdisp,p->gc_game,p->gforeground);
  p->font=XLoadFont(p->gdisp,font_graphic);
  XSetFont(p->gdisp,p->gc_game,p->font);
  XSelectInput(p->gdisp,p->win_game,KeyPressMask|ExposureMask);
  XMapRaised(p->gdisp,p->win_game);
}

object *get_player(player *p) {
  object *op=get_object();
  int i,j;

  op->x=startx,op->y=starty;
  op->face=G_CHAR;
  map[op->x][op->y]=op->face;
  op->type=PLAYER;
  op->speed=1;
  op->speed_left=1;
  op->contr=p;
  op->stats.Dex=1;
  op->stats.Str=2;
  op->stats.Con=3;
  op->stats.Wis=4;
  op->stats.Cha=5;
  op->stats.Int=6;
  op->stats.hp=7;
  op->stats.exp=8;
  op->stats.food=2000;
  for(i=WINLEFT;i<=WINRIGHT;i++)
    for(j=WINUPPER;j<=WINLOWER;j++)
      op->contr->drawn[i-WINLEFT][j-WINUPPER]='{'; /* No such character used */
  return op;
}

object *create_monster() {
  object *op=get_object();
  op->x=startx,op->y=starty;
  op->face=G_MONSTER;
  map[op->x][op->y]=op->face;
  op->type=MONSTER;
  op->speed=0.5;
  op->speed_left=1;
  return op;
}

object *get_object() {
  object *op;
  if(free_objects==NULL) {
    fprintf(stderr,"No more objects left.\n");
    leave();
  }
  op=free_objects;
  op->dirx=0,op->diry=0;
  free_objects=op->next;
  if(free_objects!=NULL)
    free_objects->prev=NULL;
  op->next=objects;
  op->prev=NULL;
  if(objects!=NULL)
    objects->prev=op;
  objects=op;
  return op;
}

free_object(object *ob) {
  if(objects==NULL) {
    fprintf(stderr,"No object to be freed.\n");
    leave();
  }

  /* First free the object from the used objects list: */
  if(ob->prev==NULL) {
    objects=ob->next;
    if(objects!=NULL)
      objects->prev=NULL;
  }
  else {
    ob->prev->next=ob->next;
    if(ob->next!=NULL)
      ob->next->prev=ob->prev;
  }
  
  /* Now link it with the free_objects list: */
  ob->prev=NULL;
  ob->next=free_objects;
  if(free_objects!=NULL)
    free_objects->prev=ob;
  free_objects=ob;
}

int distance(object *ob1,object *ob2) {
  return (ob1->x - ob2->x)*(ob1->x - ob2->x)+
         (ob1->y - ob2->y)*(ob1->y - ob2->y);
}

object *get_nearest_player(object *ob) {
  object *pl;
  int i,lastdist=0,tmp;

  pl=NULL;
  for(i=0;i<nrofplayers;i++)
    if(lastdist>(tmp=distance(players[i].ob,ob)))
      pl=ob,lastdist=tmp;
  return pl;
}

int count_free() {
  int i=0;
  object *tmp=free_objects;
  while(tmp!=NULL)
    tmp=tmp->next, i++;
  return i;
}

int count_used() {
  int i=0;
  object *tmp=objects;
  while(tmp!=NULL)
    tmp=tmp->next, i++;
  return i;
}

makemap() {
  object *op;
  int i,j;
  for(i=0;i<=mapx;i++)
    for(j=0;j<=mapy;j++)
      map[i][j]=omap[i][j];
}

int move_monster(object *op) {
  object *pl;
  pl=get_nearest_player(op);
  if(pl!=NULL&&(
     (pl->x>op->x&&move_ob(op, 1, 0))||
     (pl->x<op->x&&move_ob(op,-1, 0))||
     (pl->y<op->y&&move_ob(op,0,-1))||
     (pl->y>op->y&&move_ob(op,0,1)))) return 1;
  return 0;
}

fire(object *op,int dx,int dy) {
  object *tmp=get_object();
  tmp->dirx=dx,tmp->diry=dy;
  tmp->x=op->x,tmp->y=op->y;
  tmp->face=G_BULLET;
  tmp->speed=op->speed*1.5;
  tmp->speed_left=tmp->speed;
  tmp->type=BULLET;
  move_bullet(tmp);
  map[op->x][op->y]=op->face;
}

move_bullet(object *op) {
  map[op->x][op->y]=omap[op->x][op->y];
  op->x+=op->dirx,op->y+=op->diry;
  if(obstacle(op->x,op->y))
    free_object(op);
  else
    map[op->x][op->y]=op->face;
}

int move_ob(object *op,int dx,int dy) {
  if(obstacle(op->x+dx,op->y+dy)) return 0;
  map[op->x][op->y]=omap[op->x][op->y];
  op->x+=dx,op->y+=dy;
  map[op->x][op->y]=op->face;
  return 1;
}

draw_stats(player *pl) {
  char buff[5][MAX_BUF];
  int i;

  sprintf(buff[0],"Player: %s",pl->name);
  sprintf(buff[1],"Score: %d",pl->ob->stats.exp);
  sprintf(buff[2],"Hp: 2%d  Food: 5%d",
          pl->ob->stats.hp,pl->ob->stats.food);
  sprintf(buff[3],"Str: 2%d  Dex: 2%d  Con: 2%d",
          pl->ob->stats.Str,pl->ob->stats.Dex,pl->ob->stats.Con);
  sprintf(buff[4],"Int: 2%d  Wis: 2%d  Cha: 2%d",
          pl->ob->stats.Int,pl->ob->stats.Wis,pl->ob->stats.Cha);
  XClearWindow(pl->gdisp,pl->win_stats);
  for(i=0;i<6;i++)
    XDrawImageString(pl->gdisp,pl->win_stats,pl->gc_stats,10,i*14+10,
                     buff[i],strlen(buff[i]));
}

draw(player *pl,int x,int y) {
  int i,j,changed,ax,ay; /* ax and ay goes from 0 to max-size of arrays */
  char buff[20],tmp;
  buff[12]='\0';
  pl->last_refresh=0;
  for(j=y+WINUPPER;j<=y+WINLOWER;j++) {
    changed=0,ay=j-y-WINUPPER;
    for(i=x+WINLEFT;i<=x+WINRIGHT;i++) {
      ax=i-x-WINLEFT;
      if((tmp=(buff[ax]=(i<0||i>mapx||j<0||j>mapy)?G_BLANK:map[i][j]))
         !=pl->drawn[ax][ay])
        pl->drawn[ax][ay]=tmp,changed=1;
    }
    if(changed)
      XDrawImageString(pl->gdisp,pl->win_game,pl->gc_game,30,(j+5-y)*24+50,
                       buff,winwidth);
  }
}

main(int argc,char **argv) {
  int i;
  struct timeb tmp_time;
  unsigned short tmp_milli;

  gargc=argc, gargv=argv;
  init();
  makemap();
  for(i=0;i<nrofplayers;i++)
    draw(&players[i],players[i].ob->x,players[i].ob->y);
  while(1) {
    ftime(&tmp_time);
    tmp_milli=tmp_time.millitm;
    process_events();
    ftime(&tmp_time);
    tmp_milli+=999-tmp_time.millitm;
    if(tmp_milli<=999)
      usleep(tmp_milli);
  }
}

process_events() {
  int flag,i;
  object *op;
  player *pl;
  char text[10];

  for(flag=1;flag!=0;) {
    flag=0;
    for(op=objects;op!=NULL;op=op->next) {
      if(op->speed_left>0) {
        op->speed_left--;
        if(op->speed_left>0) flag=1; /* Another loop needed */
        switch(op->type) {
        case PLAYER:
          if (XEventsQueued(op->contr->gdisp, QueuedAfterReading)==0) {
            XFlush(op->contr->gdisp);
            break;
          }
          XNextEvent(op->contr->gdisp,&op->contr->gevent);
          switch(op->contr->gevent.type) {
          case Expose:
            if (op->contr->gevent.xexpose.count == 0) {
/*            if(op->contr->gevent.window == win_main) */
                draw(op->contr,op->x,op->y);
/*            else if(op->contr->event.window == win_stats) */
                draw_stats(op->contr);
/*            else fprintf(stderr,"Got Expose event from unknown window.\n"); */
            }
            op->speed_left++;
            break;
          case MappingNotify:
            op->speed_left++;
            XRefreshKeyboardMapping(&op->contr->gevent);
          break;
          case KeyPress:
            i=XLookupString(&op->contr->gevent,text,10,&op->contr->gkey,0);
            if(i==1) switch(text[0]) {
            case 'q': leave();
            case 'l': move_ob(op,1,0); break;
            case 'h': move_ob(op,-1,0); break;
            case 'j': move_ob(op,0,1); break;
            case 'k': move_ob(op,0,-1); break;
            case 'm': (void) create_monster(); UNDO; break;
            case 'L': UNDO; fire(op,1,0); break;
            case 'H': UNDO; fire(op,-1,0); break;
            case 'J': UNDO; fire(op,0,1); break;
            case 'K': UNDO; fire(op,0,-1); break;
            case 's':
              fprintf(stderr,"%d used, %d free.\n",count_used(),count_free());
            default:
              UNDO;
            }
          }
          break;
        case MONSTER:
          move_monster(op);
          break;
        case BULLET:
          move_bullet(op);
          break;
        }
      }
    }
  }
  for(i=0;i<nrofplayers;i++)
    draw(&players[i],players[i].ob->x,players[i].ob->y);
  for(op=objects;op!=NULL;op=op->next)
    op->speed_left+=op->speed;
}

leave() {
  int i;
  for(i=0;i<nrofplayers;i++) {
    XFreeGC(players[i].gdisp,players[i].gc_game);
    XDestroyWindow(players[i].gdisp,players[i].win_game);
    XCloseDisplay(players[i].gdisp);
    exit(0);
  }
}
